use arrayvec::ArrayString;
use rand::{seq::SliceRandom, thread_rng, CryptoRng, Rng};
use std::array;

const CHARS: &[u8] = br##"!"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_abcdefghijklmnopqrstuvwxyz{|}~"##;

bitflags::bitflags! {
    struct PasswordFlags: u8 {
        const LOWERCASE = 0b0001;
        const UPPERCASE = 0b0010;
        const NUMBER = 0b0100;
        const SPECIAL_CHARACTER = 0b1000;
    }
}

/// Returns true if the generated master password is valid.
/// It checks that it has at least one lowercase, one uppercase, one number and one punctuation char
#[inline]
fn check_generated_password<const LENGTH: usize>(password: &[u8; LENGTH]) -> bool {
    let mut flags = PasswordFlags::empty();
    for &byte in password {
        match byte {
            b'a'..=b'z' => flags |= PasswordFlags::LOWERCASE,
            b'A'..=b'Z' => flags |= PasswordFlags::UPPERCASE,
            b'0'..=b'9' => flags |= PasswordFlags::NUMBER,
            b'!'..=b'/' | b':'..=b'@' | b'['..=b'`' | b'{'..=b'~' => {
                flags |= PasswordFlags::SPECIAL_CHARACTER
            }
            _ => (),
        }
        if flags.is_all() {
            return true;
        }
    }
    false
}

#[inline]
fn generate_password<R, const LENGTH: usize>(rng: &mut R) -> ArrayString<LENGTH>
where
    R: Rng + CryptoRng,
{
    loop {
        let password = array::from_fn(|_| *CHARS.choose(rng).unwrap());
        if check_generated_password(&password) {
            return ArrayString::from_byte_string(&password).unwrap();
        }
    }
}

/// Continuously generates the password until it passes the checks
#[inline]
pub fn generate_passwords<const AMOUNT: usize, const LENGTH: usize>(
) -> [ArrayString<LENGTH>; AMOUNT] {
    let mut rng = thread_rng();
    array::from_fn(|_| generate_password(&mut rng))
}

#[inline]
pub fn check_master_pass(password: &str) -> bool {
    if password.chars().count() < 8 {
        return false;
    }

    let mut flags = PasswordFlags::empty();
    for char in password.chars() {
        if char.is_lowercase() {
            flags |= PasswordFlags::LOWERCASE
        } else if char.is_uppercase() {
            flags |= PasswordFlags::UPPERCASE
        } else if char.is_ascii_digit() {
            flags |= PasswordFlags::NUMBER;
        } else if char.is_ascii_punctuation() {
            flags |= PasswordFlags::SPECIAL_CHARACTER
        }

        if flags.is_all() {
            return true;
        }
    }
    false
}
